from __future__ import annotations

from dataclasses import dataclass
from typing import Tuple, List, Callable

import js
from js import document, console, ResizeObserver
from pyodide.ffi import create_proxy

import wwwpy.remote.component as wpc
from wwwpy.remote import dict_to_js
from typing import NamedTuple
import logging

logger = logging.getLogger(__name__)


class Geometry(NamedTuple):
    left: int
    top: int
    width: int
    height: int


class WindowComponent(wpc.Component, tag_name='wwwpy-window'):
    _window_div: wpc.HTMLElement = wpc.element()
    _title_div: wpc.HTMLElement = wpc.element()
    _title: js.HTMLElement = wpc.element()
    _close: js.HTMLElement = wpc.element()

    client_x: int
    client_y: int
    css_border = 2  # 1px border on each side, so we need to subtract 2px from width and height
    geometry_change_listeners: List[Callable[[], None]] = []

    @property
    def title(self) -> str:
        return self._title.innerHTML

    @title.setter
    def title(self, value: str):
        self._title.innerHTML = value

    @property
    def closable(self) -> bool:
        return self._close.style.display != 'none'

    @closable.setter
    def closable(self, value: bool):
        self._close.style.display = 'block' if value else 'none'

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
<style>
.window {
  z-index: 100000;  
  background-color: black;
  border: 1px solid #d3d3d3;
  resize: both;  
  overflow: hidden;
  position: absolute;  /* Changed from 'relative' to 'absolute' */
  display: flex;
  flex-direction: column;
}

.window-title {
  padding: 10px;
  cursor: move;
  z-index: 1001;
  background-color: #2196F3;
  color: #fff;
  touch-action: none;
}

.window-body {
  overflow: auto;
}
</style>        
<div data-name="_window_div" class='window'>
   <div data-name="_title_div" class='window-title' style="display: flex; justify-content: center; align-items: center;">
    <slot name='title' data-name="_title" style='text-align: center'>slot=title</slot>
    <button data-name="_close" style="margin-left: auto; cursor: pointer;">X</button>
</div>
   <div class='window-body'>
        <slot>slot=default</slot>
    </div>    
</div> 
"""
        # we don't rely on self._window_div.offsetLeft/offsetTop for reading x/y location
        # because they need rendering to be up to date
        self._window_div.style.top = '10px'
        self._window_div.style.left = '20px'
        self._win_move(0, 0)

        self._pointermove_proxy = create_proxy(self._pointermove)
        self._pointerup_proxy = create_proxy(self._pointerup)

        def on_resize(entries, observer):
            self._on_geometry_change()

        resize_observer = ResizeObserver.new(create_proxy(on_resize))
        resize_observer.observe(self._window_div)

    def _close__pointerdown(self, e: js.PointerEvent):
        e.preventDefault()
        e.stopPropagation()
        self.element.remove()

    def _on_geometry_change(self):
        for listener in self.geometry_change_listeners:
            listener()

    def _title_div__pointerdown(self, e: js.PointerEvent):
        self._move_start(e)

    def _move_start(self, e: js.PointerEvent):
        e.preventDefault()
        self._update_client_xy(e)

        self._title_div.setPointerCapture(e.pointerId)

        self._title_div.addEventListener('pointermove', self._pointermove_proxy)
        self._title_div.addEventListener('pointerup', self._pointerup_proxy)

    def _update_client_xy(self, e):
        self.client_x = e.clientX
        self.client_y = e.clientY

    def _pointermove(self, event: js.PointerEvent):
        self._win_move(self.client_x - event.clientX, self.client_y - event.clientY)
        self._update_client_xy(event)

    def _win_move(self, delta_x, delta_y):
        current_left = float(self._window_div.style.left.rstrip('px'))
        current_top = float(self._window_div.style.top.rstrip('px'))
        new_left = current_left - delta_x
        new_top = current_top - delta_y
        self.set_position(f'{new_left}px', f'{new_top}px')
        self._on_geometry_change()

    def _pointerup(self, event: js.PointerEvent):
        self._title_div.removeEventListener('pointermove', self._pointermove_proxy)
        self._title_div.removeEventListener('pointerup', self._pointerup_proxy)
        self._title_div.releasePointerCapture(event.pointerId)

    def geometry(self) -> Geometry:
        t = self._window_div
        res = Geometry(
            t.offsetLeft,
            t.offsetTop,
            t.offsetWidth - self.css_border,
            t.offsetHeight - self.css_border
        )
        return res

    def set_geometry(self, geometry_tuple: Geometry):
        left, top, width, height = geometry_tuple
        self.set_position(f"{left}px", f"{top}px")
        self.set_size(f"{height}px", f"{width}px")

    def set_position(self, left: str | None = None, top: str | None = None):
        if top:
            top_check = float(top.removesuffix('px'))
            if top_check < 0:
                top = '0px'
            self._window_div.style.top = top
        if left:
            g = self.geometry()
            left_check = float(left.removesuffix('px'))
            if (left_check + g.width) > 30:
                self._window_div.style.left = left
        logger.debug(f'set_position top={top} left={left}')

    def set_size(self, height: str | None = None, width: str | None = None):
        if height:
            self._window_div.style.height = height
        if width:
            self._window_div.style.width = width


def new_window(title: str, closable=True) -> WindowComponent:
    win = WindowComponent()
    win.closable = closable
    win.title = title
    return win




from __future__ import annotations

import logging

import js

import wwwpy.remote.component as wpc
from wwwpy.common.asynclib import create_task_safe
from wwwpy.remote import dict_to_js
from wwwpy.server.designer import rpc
from . import quickstart_ui
from .quickstart_ui import QuickstartUI
from .toolbox import ToolboxComponent

logger = logging.getLogger(__name__)


def show():
    x = DevModeComponent.instance
    js.document.body.append(DevModeComponent.instance.element)


class classproperty(property):
    def __get__(self, cls, owner):
        return classmethod(self.fget).__get__(None, owner)()


class DevModeComponent(wpc.Component, tag_name='wwwpy-dev-mode-component'):
    toolbox: ToolboxComponent = wpc.element()
    quickstart: QuickstartUI | None = None
    _instance: DevModeComponent

    @classproperty
    def instance(cls) -> DevModeComponent:  # noqa
        try:
            _i = cls._instance
            logger.debug(f'instance found: {_i}')
        except AttributeError:
            _i = DevModeComponent()
            cls._instance = _i
            logger.debug(f'instance created: {_i}')
        return _i

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
<wwwpy-toolbox data-name="toolbox"></wwwpy-toolbox>        
        """

    async def _connectedCallback_async(self):
        if await rpc.quickstart_possible():
            self.quickstart = quickstart_ui.create()

            def _on_done():
                self.toolbox.visible = True

            self.quickstart.on_done = lambda *_: _on_done()
            self.element.shadowRoot.append(self.quickstart.window.element)
            self.toolbox.visible = False
        else:
            self.toolbox.visible = True

    def connectedCallback(self):
        create_task_safe(self._connectedCallback_async())



import subprocess
from pathlib import Path

wd = Path(__file__).parent.parent.parent

# uv pip install granian
subprocess.run(['granian', '--interface', 'asgi', 'wwwpy.asgi.echo_handler:app'], cwd=wd)



from __future__ import annotations

from typing import Callable, Union, Awaitable


# PUBLIC-API
def dict_to_js(o):
    import js
    import pyodide
    return pyodide.ffi.to_js(o, dict_converter=js.Object.fromEntries)


async def micropip_install(package):
    from js import pyodide
    await pyodide.loadPackage('micropip')
    import micropip
    await micropip.install([package])


# def set_timeout(callback: Callable[[], Union[None, Awaitable[None]]], timeout_millis: int | None = 0):
#     from pyodide.ffi import create_once_callable
#     from js import window
#     window.setTimeout(create_once_callable(callback), timeout_millis)


def dict_to_py(js_obj):
    import js
    py_dict = {}
    # Iterate through properties, including those in the prototype chain
    current = js_obj
    while current:
        for prop in js.Object.getOwnPropertyNames(current):
            try:
                value = getattr(js_obj, prop)
                if not callable(value):  # Skip methods
                    py_dict[prop] = value
            except Exception:
                # Ignore properties that might throw errors when accessed
                continue
        current = js.Object.getPrototypeOf(current)

    return py_dict



from __future__ import annotations

from js import HTMLElement, console

from wwwpy.common.designer.element_editor import ElementEditor, EventEditor
from wwwpy.common.designer.element_library import ElementDef
from wwwpy.common.designer.element_path import ElementPath
from wwwpy.remote.designer import element_path
from wwwpy.server.designer import rpc


async def _rpc_save(el_path: ElementPath, new_source: str):
    await rpc.write_module_file(el_path.class_module, new_source)


def info_link(href):
    # open in a new page
    # <a href='https://www.google.com' target='_blank'><svg class="help-icon"><use href="#help-icon"/></svg></a>
    # language=html
    return f"""<wwwpy-help-icon href="{href}"></wwwpy-help-icon>"""

    return (f'<a href="{href}" style="text-decoration: none" target="_blank">'
            f'<svg class="help-icon"><use href="#help-icon"/></svg></a>')


def _element_lbl(element: HTMLElement) -> str:
    ep = element_path.element_path(element)
    console.log(f'element_path={ep}')
    return _element_path_lbl(ep) if ep else 'No element path'


def _element_path_lbl(ep: ElementPath | None) -> str:
    from wwwpy.common import modlib
    class_file_path = modlib._find_module_path(ep.class_module)
    cfp = '' if not class_file_path else f'{class_file_path.name}::'
    lbl = '' if not ep.data_name else f' ({ep.data_name})'
    msg = f'{ep.tag_name} element {lbl} in {cfp}{ep.class_name}'
    return msg


async def _log_event(element_editor: ElementEditor, event_editor: EventEditor):
    ep = element_editor.element_path
    ee = ElementEditor(ep, element_editor.element_def)
    ev = ee.events.get(event_editor.definition.name)
    message = f'\nClick below to locate the event handler {ev.method.name}'
    await rpc.print_module_line(ep.class_module, message, ev.method.code_lineno)


def _help_button(element_def: ElementDef) -> str:
    help_url = element_def.help.url
    help_button = '' if len(help_url) == 0 else info_link(element_def.help.url)
    return help_button


async def _on_error(message, source, lineno, colno, error):
    await rpc.on_error(message, source, lineno, colno, str(error))


async def _on_unhandledrejection(event):
    try:
        await rpc.on_unhandledrejection(f'{event.reason}')
    except Exception as ex:
        # trap and send to the console, otherwise we will trigger this handler again going into an infinite loop
        console.error(f'Error in on_unhandledrejection: {ex}')


def _help_url(help_str: str) -> str:
    return f'https://wwwpy.dev/help/{help_str}.html'



from __future__ import annotations

from dataclasses import dataclass
from typing import Callable, List

from wwwpy.common.property_monitor import Monitor, PropertyChanged, set_origin, monitor_changes, Monitorable, \
    get_monitor_or_create


@dataclass
class TargetOriginEvent:
    origin_target: TargetAdapter
    value: any


# todo rename to BindingTarget or Bindable or BindableTarget?
class TargetAdapter(Monitorable):

    def __init__(self):
        super().__init__()

    def set_target_value(self, value):
        pass

    def get_target_value(self):
        pass


class Binding:

    def __init__(self, source, attr_name, target_adapter: TargetAdapter):
        super().__init__()
        self.source = source
        self.attr_name = attr_name
        self.target_adapter = target_adapter
        target_adapter.monitor_object.listeners.append(self._on_target_changes)
        get_monitor_or_create(source).add_attribute_listener(attr_name, self._on_source_changes)

    def apply_binding(self):
        self.target_adapter.set_target_value(getattr(self.source, self.attr_name))

    def _on_target_changes(self, events: List[PropertyChanged]):
        event = events[-1]
        with set_origin(self.source, self):
            setattr(self.source, self.attr_name, event.new_value)

    def _on_source_changes(self, events: List[PropertyChanged]):
        event = events[-1]  # todo we should filter only for the attr_name!
        if event.origin == self:
            return
        with set_origin(self.target_adapter, self):
            self.target_adapter.set_target_value(event.new_value)



from __future__ import annotations

from collections import UserList
from dataclasses import dataclass, field
from .html_parser_mod import HTMLParser
from typing import Tuple, Dict, List


@dataclass
class CstAttribute:
    name: str
    value: str | None
    name_span: Tuple[int, int]
    value_span: Tuple[int, int] | None
    """The span of the attribute value in the HTML string including the quotes char."""
    child_index: int
    """The index of the attribute in the parent node."""


def cst_attribute_dict(*attributes: CstAttribute) -> Dict[str, CstAttribute]:
    return {attr.name: attr for attr in attributes}


@dataclass
class CstNode:
    tag_name: str
    span: Tuple[int, int]
    """The span of the node in the HTML string, including the `<` and `>` characters."""
    attr_span: Tuple[int, int]
    """The span of the attributes part. If there is no span the tuple will contain the same value."""
    content_span: Tuple[int, int] | None = None
    """The span of the content part. If there is no content the tuple will contain the same value. 
    If the node is a void tag, this will be None."""
    children: CstTree = field(default_factory=lambda: CstTree())
    attributes_list: List[CstAttribute] = field(default_factory=list)
    html: str = field(repr=False, compare=False, default='')
    content: str | None = field(repr=False, compare=False, default=None)
    parent: CstNode | None = field(repr=False, compare=False, default=None)
    level: int = field(repr=False, compare=False, default=0)
    child_index: int = field(repr=False, compare=False, default=0)

    def __post_init__(self):
        self._attributes_dict = None

    @property
    def attributes(self) -> Dict[str, str]:
        if self._attributes_dict is None:
            self._attributes_dict = {attr.name: attr.value for attr in self.attributes_list}
        return self._attributes_dict

    def cst_attribute(self, name: str) -> CstAttribute | None:
        for attr in self.attributes_list:
            if attr.name == name:
                return attr
        return None


class CstTree(UserList[CstNode]):
    def traverse(self, indexes: list[int]) -> CstNode | None:
        children = self
        node = None
        for i in indexes:
            if i < 0 or i >= len(children):
                return None
            node = children[i]
            children = node.children
        return node


def html_to_tree(html: str) -> CstTree:
    parser = _PositionalHTMLParser(html)
    parse = parser.parse()
    _complete_tree_data(html, parse)
    return parse


class _PositionalHTMLParser(HTMLParser):
    void_tags = {'area', 'base', 'br', 'col', 'embed', 'hr', 'img', 'input', 'link', 'meta', 'source', 'track', 'wbr'}

    def __init__(self, html: str):
        super().__init__()
        self.html = html
        self.nodes = CstTree()
        self.stack = []
        self.current_pos = 0

    def handle_starttag_extended(self, tag, attrs, attrs_extended, autoclosing):
        start_pos = self.current_pos

        def _cst_attr(name, v, idx) -> CstAttribute:
            return CstAttribute(name, v['value'], v['name_span'], v['value_span'], idx)

        text = self.get_starttag_text()
        end_displ = 2 if autoclosing else 1
        attr_span = (start_pos + len(tag) + 1, self.current_pos + len(text) - end_displ)
        node = CstNode(
            tag_name=tag,
            span=(start_pos, None),
            attr_span=attr_span,
            content_span=None if autoclosing else (attr_span[1] + 1, None),
            attributes_list=[_cst_attr(name, v, idx) for idx, (name, v) in enumerate(attrs_extended.items())],
        )

        if self.stack:
            self.stack[-1].children.append(node)

        if tag not in self.void_tags:
            self.stack.append(node)
        else:
            node.span = (start_pos, self.current_pos + len(text))
            node.content_span = None
            if not self.stack:
                self.nodes.append(node)

        self.current_pos += len(text)

    def handle_endtag(self, tag):
        if tag in self.void_tags:  # a void tag with end tag
            return
        if not self.stack:
            return

        node = self.stack.pop()
        span_end = self.current_pos + len(tag) + 3  # +3 for </ and >
        node.span = (node.span[0], span_end)
        if node.content_span:
            node.content_span = (node.content_span[0], self.current_pos)
        if not self.stack:
            self.nodes.append(node)

        self.current_pos += len(tag) + 3

    def handle_data(self, data):
        self.current_pos += len(data)

    def parse(self):
        self.feed(self.html)
        return self.nodes


def _complete_tree_data(html: str, tree: CstTree, parent: CstNode | None = None, level: int = 0):
    for idx, node in enumerate(tree):
        node.parent = parent
        node.level = level
        node.child_index = idx
        start, end = node.span
        node.html = html[start:end]
        c_sp = node.content_span
        node.content = html[c_sp[0]:c_sp[1]] if c_sp else None
        _complete_tree_data(html, node.children, node, level + 1)



"""This module contains the HTML string manipulator functions."""
from __future__ import annotations

from enum import Enum
from html import escape
from wwwpy.common.designer.html_locator import NodePath, IndexPath, check_node_path
from wwwpy.common.designer import html_locator


class Position(str, Enum):
    inside = 'inside'
    beforebegin = 'beforebegin'
    afterend = 'afterend'


def html_add(html: str, add: str, node_path: NodePath, position: Position) -> str:
    """This function adds an HTML piece to the specified position in the HTML string."""

    start, end = html_locator.locate_span(html, node_path)

    index = start if position == Position.beforebegin else end

    return html[:index] + add + html[index:]


def html_add_indexed(html: str, add: str, index_path: IndexPath, position: Position) -> str:
    check_node_path(index_path)
    """This function adds an HTML piece to the specified position in the HTML string."""

    start, end = html_locator.locate_span_indexed(html, index_path)

    index = start if position == Position.beforebegin else end

    return html[:index] + add + html[index:]


def html_edit(html: str, edit: str, node_path: NodePath) -> str:
    """This function edits the HTML string at the specified path."""
    start, end = html_locator.locate_span(html, node_path)

    return html[:start] + edit + html[end:]


def html_edit_indexed(html: str, edit: str, index_path: NodePath) -> str:
    check_node_path(index_path)
    """This function edits the HTML string at the specified path."""
    start, end = html_locator.locate_span_indexed(html, index_path)

    return html[:start] + edit + html[end:]


def html_attribute_set(html: str, node_path: NodePath, attr_name: str, attr_value: str | None) -> str:
    """This function sets an attribute of the specified node in the HTML string.
    When the value is None, the attribute value is remove entirely"""

    node = html_locator.locate_node(html, node_path)
    if node is None:
        print(f'node not found at path={node_path} in html=```{html}```')
        return html

    cst_attr = node.cst_attribute(attr_name)
    attr_present = cst_attr is not None
    value_present = attr_present and cst_attr.value_span is not None
    spacer = ' ' if not attr_present else ''
    sep = '"' if not value_present else html[cst_attr.value_span[0]]

    value_part = '' if attr_value is None else '=' + sep + escape(attr_value) + sep

    x = cst_attr.name_span[0] if attr_present else node.attr_span[1]
    y = cst_attr.value_span[1] if value_present else (x if not attr_present else cst_attr.name_span[1])
    left = html[:x]
    right = html[y:]

    return left + spacer + attr_name + value_part + right


def html_attribute_remove(html: str, node_path: NodePath, attr_name: str) -> str:
    node = html_locator.locate_node(html, node_path)
    if node is None:
        return html

    cst_attr = node.cst_attribute(attr_name)
    if cst_attr is None:
        return html

    if len(node.attributes) == 1:
        x = node.attr_span[0]
        y = node.attr_span[1]
        return html[:x] + html[y:]


    value_present = cst_attr.value_span is not None
    x = cst_attr.name_span[0]
    y = cst_attr.value_span[1] if value_present else cst_attr.name_span[1]

    left = html[:x]
    right = html[y:]

    left = left.rstrip()
    right = right.lstrip()

    is_leftmost = cst_attr.child_index == 0
    space = ' ' if is_leftmost else ''

    return left + space + right


def html_content_set(html: str, node_path: NodePath, value: str) -> str | None:
    """This function sets the content of the specified node in the HTML string."""

    node = html_locator.locate_node(html, node_path)
    if node is None:
        raise Exception(f'node not found at path={node_path} in html=```{html}```')

    if node.content_span is None:
        raise Exception(f'node is a void tag at path={node_path} in html=```{html}```')

    x, y = node.content_span
    left = html[:x]
    right = html[y:]
    return left + value + right



from __future__ import annotations

from typing import TypeVar, Union, Type

from wwwpy.common.rpc import serialization

T = TypeVar('T')


class EncoderDecoderType:
    buffer_type: Union[bytes, str]


class Encoder(EncoderDecoderType):
    buffer: bytes | str

    def encode(self, obj: any, cls: Type[T]):
        """Add the object to the encoder"""


class Decoder(EncoderDecoderType):
    def decode(self, cls: Type[T]) -> T:
        """Get the next object of the given type"""


class EncoderDecoder:
    def decoder(self, buffer: str | bytes) -> Decoder: raise NotImplementedError

    def encoder(self) -> Encoder: raise NotImplementedError


class JsonEncoder(Encoder):
    def __init__(self, sep: str):
        self._sep = sep
        self._buffer = []

    def encode(self, obj: any, cls: Type[T]):
        self._buffer.append(serialization.to_json(obj, cls))

    @property
    def buffer(self) -> str:
        return self._sep.join(self._buffer)


class JsonDecoder(Decoder):
    def __init__(self, buffer: str, sep: str):
        self._buffer = iter(buffer.split(sep))

    def decode(self, cls: Type[T]) -> T:
        item = next(self._buffer)
        return serialization.from_json(item, cls)


class JsonEncoderDecoder(EncoderDecoder):

    def __init__(self):
        self._sep = '\n'

    def decoder(self, buffer: str | bytes) -> Decoder:
        return JsonDecoder(buffer, sep=self._sep)

    def encoder(self) -> Encoder:
        return JsonEncoder(sep=self._sep)



from __future__ import annotations

import traceback
from contextlib import contextmanager
from dataclasses import dataclass
from types import FunctionType

from wwwpy.common.rpc2.encoder_decoder import EncoderDecoder
from wwwpy.common.rpc2.skeleton import Skeleton
from wwwpy.common.rpc2.transport import Transport
from wwwpy.common.rpc2.typed_function import get_typed_function, TypedFunction
from wwwpy.unasync import unasync


def _get_args_types(func): ...


def _get_return_type(func): ...


class DefaultSkeleton(Skeleton):
    def __init__(self, transport: Transport, encdec: EncoderDecoder, allowed_modules: set[str]):
        self._transport = transport
        self._encdec = encdec
        self._allowed_modules = allowed_modules

    def invoke_tobe_fixed(self):
        recv_buffer = self._transport.recv_sync()
        args, func, target_function = self._decode_request(recv_buffer)

        if target_function.is_coroutine:
            @unasync
            async def func_async(*args):
                with _catch() as r:
                    r.value = await func(*args)
                return r

            r = func_async(*args)
        else:
            with _catch() as r:
                r.value = func(*args)

        send_buffer = self._encode_result(target_function, r)
        self._transport.send_sync(send_buffer)

    def invoke_sync(self):
        # raise Exception('Not implemented - see invoke_tobe_fixed')
        recv_buffer = self._transport.recv_sync()
        args, func, target_function = self._decode_request(recv_buffer)

        with _catch() as r:
            r.value = func(*args)

        send_buffer = self._encode_result(target_function, r)
        self._transport.send_sync(send_buffer)

    async def invoke_async(self):
        # raise Exception('Not implemented - see invoke_tobe_fixed')
        recv_buffer = await self._transport.recv_async()
        args, func, target_function = self._decode_request(recv_buffer)

        with _catch() as r:
            r.value = await func(*args)

        send_buffer = self._encode_result(target_function, r)
        await self._transport.send_async(send_buffer)

    def _decode_request(self, recv_buffer) -> tuple[list[any], FunctionType, TypedFunction]:
        decoder = self._encdec.decoder(recv_buffer)
        module_name = decoder.decode(str)
        if module_name not in self._allowed_modules:
            raise Exception(f'Not allowed module: {module_name}')
        func_name = decoder.decode(str)
        import importlib
        module = importlib.import_module(module_name)
        func = getattr(module, func_name)
        target_function = get_typed_function(func)
        args = []
        for arg_type in target_function.args_types:
            args.append(decoder.decode(arg_type))
        return args, func, target_function

    def _encode_result(self, target_function, result: _Result):
        encoder = self._encdec.encoder()
        if result.exception_str:
            encoder.encode('ex', str)
            encoder.encode(result.exception_str, str)
        else:
            encoder.encode('ok', str)
            encoder.encode(result.value, target_function.return_type)
        send_buffer = encoder.buffer
        return send_buffer


@dataclass
class _Result:
    value: any = None
    exception_str: str | None = None


@contextmanager
def _catch(*args, **kwds):
    r = _Result()
    try:
        yield r
    except Exception as e:
        r.exception_str = traceback.format_exc()



from __future__ import annotations

import logging
from dataclasses import dataclass
from pathlib import Path

import libcst as cst

from wwwpy.common.designer import code_info, html_parser, html_locator
from wwwpy.common.designer.code_info import Attribute
from wwwpy.common.designer.code_strings import html_string_edit
from wwwpy.common.designer.element_library import ElementDef
from wwwpy.common.designer.html_edit import Position, html_add_indexed
from wwwpy.common.designer.html_locator import NodePath, IndexPath

logger = logging.getLogger(__name__)


def add_class_attribute(source_code: str, class_name: str, attr_info: Attribute) -> str:
    source_code_imp = ensure_imports(source_code)
    module = cst.parse_module(source_code_imp)
    transformer = _AddFieldToClassTransformer(class_name, attr_info)
    modified_tree = module.visit(transformer)

    return modified_tree.code


def rename_class_attribute(source_code: str, class_name: str, old_attr_name: str, new_attr_name: str):
    source_code_imp = ensure_imports(source_code)
    # tree1 = tree0.visit(_RenameFieldInClassTransformer(class_name, old_attr_name, new_attr_name))
    tree0 = (cst.parse_module(source_code_imp).
             visit(_RenameMethodEventsInClassTransformer(class_name, old_attr_name, new_attr_name)))
    code_rope = _rope_rename_class_attribute(tree0.code, class_name, old_attr_name, new_attr_name)

    return code_rope


def _rope_rename_class_attribute(source_code: str, class_name: str, old_attr_name: str, new_attr_name: str):
    import tempfile
    import shutil

    temp_dir = Path(tempfile.mkdtemp())
    try:
        temp_file = temp_dir / 'temp.py'
        temp_file.write_text(source_code)
        return _rope_rename_class_attribute_in_project(temp_dir, temp_file, class_name, old_attr_name, new_attr_name)
    finally:
        shutil.rmtree(temp_dir)


def _rope_rename_class_attribute_in_project(project_dir: Path, file_path: Path, class_name: str, old_attr_name: str,
                                            new_attr_name: str):
    from rope.base.project import Project
    from rope.base.libutils import path_to_resource
    from rope.refactor.rename import Rename
    import re
    project = Project(str(project_dir))
    resource = path_to_resource(project, str(file_path))
    orig_code = resource.read()
    pattern = rf'class\s+{class_name}\b[\s\S]*?\b{old_attr_name}\b'
    match = re.search(pattern, orig_code)
    if not match:
        return orig_code
    name_offset = match.start() + match.group(0).rfind(old_attr_name)
    rename = Rename(project, resource, name_offset)
    changes = rename.get_changes(new_attr_name)
    project.do(changes)
    project.close()
    return resource.read()


@dataclass
class AddResult:
    source_code: str
    node_path: NodePath


@dataclass
class AddFailed:
    exception: Exception
    exception_report_str: str
    exception_report_b64: str


@dataclass
class AddComponentExceptionReport:
    stack_trace: str
    source_code_orig: str
    class_name: str
    tag_name: str
    index_path: IndexPath
    position: Position


def add_component(source_code: str, class_name: str, comp_def: ElementDef, index_path: IndexPath,
                  position: Position) -> AddResult | AddFailed:
    source_code_orig = source_code
    try:
        source_code = ensure_imports(source_code)
        class_info = code_info.class_info(source_code, class_name)
        if class_info is None:
            print(f'Class {class_name} not found inside source ```{source_code}```')
            return None

        attr_name = class_info.next_attribute_name(comp_def.tag_name)
        named_html = comp_def.new_html(attr_name)

        source1 = add_class_attribute(source_code, class_name,
                                      Attribute(attr_name, comp_def.python_type, 'wpc.element()'))

        def manipulate_html(html):
            add = html_add_indexed(html, named_html, index_path, position)
            return add

        source2 = html_string_edit(source1, class_name, manipulate_html)
        new_tree = html_parser.html_to_tree(source2)
        displacement = 0 if position == Position.beforebegin else 1
        indexes = index_path[0:-1] + [index_path[-1] + displacement]
        new_node_path = html_locator.tree_to_path(new_tree, indexes)
        result = AddResult(source2, new_node_path)
    except Exception as e:
        import traceback
        from wwwpy.common.rpc import serialization
        logger.error(f'Error adding component: {e}')
        logger.exception('Full exception report')

        exception_report = AddComponentExceptionReport(traceback.format_exc(), source_code_orig, class_name,
                                                       comp_def.tag_name, index_path, position)
        exception_report_str = serialization.to_json(exception_report, AddComponentExceptionReport)
        logger.error(f'Exception report str:\n{"=" * 20}\n{exception_report_str}\n{"=" * 20}')
        from wwwpy.common.files import str_gzip_base64
        exception_report_b64 = str_gzip_base64(exception_report_str)
        logger.error(f'Exception report b64:\n{"=" * 20}\n{exception_report_b64}\n{"=" * 20}')
        e.exception_report_b64 = exception_report_b64
        return AddFailed(e, exception_report_str, exception_report_b64)
    return result


class _RenameFieldInClassTransformer(cst.CSTTransformer):
    def __init__(self, class_name, old_field_name, new_field_name):
        super().__init__()
        self.class_name = class_name
        self.old_field_name = old_field_name
        self.new_field_name = new_field_name

    def leave_ClassDef(self, original_node, updated_node):
        if original_node.name.value != self.class_name:
            return original_node
        new_body = []
        for item in updated_node.body.body:
            if isinstance(item, cst.SimpleStatementLine) and isinstance(item.body[0], cst.AnnAssign):
                ann_assign = item.body[0]
                if ann_assign.target.value == self.old_field_name:
                    new_body.append(
                        item.with_changes(body=[ann_assign.with_changes(target=cst.Name(self.new_field_name))]))
                else:
                    new_body.append(item)
            else:
                new_body.append(item)
        return updated_node.with_changes(body=updated_node.body.with_changes(body=new_body))


# class _RenameMethodEventsInClassTransformer(cst.CSTTransformer):
#     """This transformer given an old_file_name='btn1' and new_field_name='btn2' will rename all the methods that
#      starts with 'btn1__' to 'btn2__'"""
#     def __init__(self, class_name, old_field_name, new_field_name):
#         super().__init__()

class _RenameMethodEventsInClassTransformer(cst.CSTTransformer):
    def __init__(self, class_name, old_field_name, new_field_name):
        super().__init__()
        self.class_name = class_name
        self.old_field_name = old_field_name
        self.new_field_name = new_field_name

    def leave_ClassDef(self, original_node, updated_node):
        if original_node.name.value != self.class_name:
            return original_node
        new_body = []
        for item in updated_node.body.body:
            if isinstance(item, cst.FunctionDef):
                if item.name.value.startswith(self.old_field_name + '__'):
                    new_body.append(item.with_changes(
                        name=cst.Name(item.name.value.replace(self.old_field_name, self.new_field_name))))
                else:
                    new_body.append(item)
            else:
                new_body.append(item)
        return updated_node.with_changes(body=updated_node.body.with_changes(body=new_body))


class _AddFieldToClassTransformer(cst.CSTTransformer):
    def __init__(self, class_name, new_field: Attribute):
        super().__init__()
        self.class_name = class_name
        self.new_field = new_field

    def leave_ClassDef(self, original_node, updated_node):
        if original_node.name.value != self.class_name:
            return original_node
        # Check if the type is a composite name (contains a dot)
        if '.' in self.new_field.type:
            base_name, attr_name = self.new_field.type.rsplit('.', 1)
            annotation = cst.Annotation(cst.Attribute(value=cst.Name(base_name), attr=cst.Name(attr_name)))
        else:
            annotation = cst.Annotation(cst.Name(self.new_field.type))

        new_field_node = cst.SimpleStatementLine([
            cst.AnnAssign(
                target=cst.Name(self.new_field.name),
                annotation=annotation,
                value=None if self.new_field.default is None else cst.parse_expression(self.new_field.default)
            )
        ])

        # Find the position to insert the new attribute
        last_assign_index = 0
        for i, item in enumerate(updated_node.body.body):
            if isinstance(item, cst.SimpleStatementLine) and isinstance(item.body[0], cst.AnnAssign):
                last_assign_index = i + 1

        new_body = list(updated_node.body.body)
        new_body.insert(last_assign_index, new_field_node)

        return updated_node.with_changes(body=updated_node.body.with_changes(body=new_body))


def add_method(source_code: str, class_name: str, method_name: str, method_args: str,
               instructions: str = 'pass') -> str:
    source_code_imp = ensure_imports(source_code)
    module = cst.parse_module(source_code_imp)
    transformer = _AddMethodToClassTransformer(True, class_name, method_name, 'self, ' + method_args, instructions)
    modified_tree = module.visit(transformer)
    return modified_tree.code


class _AddMethodToClassTransformer(cst.CSTTransformer):
    def __init__(self, async_def: bool, class_name: str, method_name: str, method_args: str, instructions: str):
        super().__init__()
        self.async_def = async_def
        self.class_name = class_name
        self.method_name = method_name
        self.method_args = method_args
        self.instructions = instructions

    def leave_ClassDef(self, original_node, updated_node):
        if original_node.name.value != self.class_name:
            return original_node

        parsed_instructions = cst.parse_module(self.instructions).body

        new_method_node = cst.FunctionDef(
            name=cst.Name(self.method_name),
            params=cst.Parameters(
                params=[cst.Param(name=cst.Name(arg.strip())) for arg in self.method_args.split(',') if arg.strip()]
            ),
            body=cst.IndentedBlock(body=parsed_instructions),
            asynchronous=cst.Asynchronous() if self.async_def else None
        )

        new_body = list(updated_node.body.body)
        new_body.append(cst.EmptyLine())
        new_body.append(new_method_node)
        new_body.append(cst.EmptyLine())

        return updated_node.with_changes(body=updated_node.body.with_changes(body=new_body))


def ensure_imports(source_code: str) -> str:
    required_imports = [
        'import inspect',
        'import logging',
        'import js',
        'import wwwpy.remote.component as wpc',
    ]

    def _remove_comment_if_present(line) -> str:
        line = line.strip()
        if '#' in line:
            line = line[:line.index('#')]
        return line.strip()

    existing_imports = set(_remove_comment_if_present(line) for line in source_code.split('\n')
                           if line.strip().startswith('import'))

    for imp in required_imports:
        if imp not in existing_imports:
            source_code = imp + '\n' + source_code

    return source_code



from __future__ import annotations

import argparse
import os
from pathlib import Path
from typing import NamedTuple, Optional, Sequence

from wwwpy.server.convention import start_default


class Arguments(NamedTuple):
    directory: Path
    port: int
    dev: bool


def parse_arguments(args: Optional[Sequence[str]] = None) -> Arguments:
    parser = argparse.ArgumentParser(prog='wwwpy')
    parser.add_argument('dev', nargs='?', const=True, default=False,
                        help="Run in development mode")
    parser.add_argument('--directory', '-d', default=os.getcwd(),
                        help='set the root path for the project (default: current directory)')
    parser.add_argument('--port', type=int, default=8000,
                        help='bind to this port (default: 8000)')

    parsed_args = parser.parse_args(args)
    return Arguments(
        directory=Path(parsed_args.directory).absolute(),
        port=parsed_args.port,
        dev=bool(parsed_args.dev)
    )


def _open_browser(args, settings):
    import webbrowser
    if settings.open_url_code:
        loc = {'url': f'http://localhost:{args.port}', args: args}
        exec(settings.open_url_code, loc, loc)
    else:
        webbrowser.open(f'http://localhost:{args.port}')


def main():
    import wwwpy
    print(f'Starting wwwpy v{wwwpy.__version__}')
    args = parse_arguments()
    project = start_default(args.directory, args.port, dev_mode=args.dev)
    _open_browser(args, project.settings)
    try:
        from wwwpy.webserver import wait_forever
        wait_forever()
    except KeyboardInterrupt:
        pass


if __name__ == '__main__':
    main()



from wwwpy.common.databind.databind import TargetAdapter
import js

from pyodide.ffi import create_proxy

from wwwpy.common.property_monitor import PropertyChanged

# rename to HTMLInputTargetAdapter or HTMLInputTarget
class InputTargetAdapter(TargetAdapter):
    def __init__(self, inp: js.HTMLInputElement):
        super().__init__()
        self.input = inp
        self.input.addEventListener('input', create_proxy(self._new_input_event))

    def set_target_value(self, value):
        self.input.value = value

    def get_target_value(self):
        return self.input.value

    def _new_input_event(self, event):
        self.monitor_object.notify([PropertyChanged(self, '', None, self.input.value)])



# wwwpy/pytest_plugin.py
import importlib.util
import inspect
from pathlib import Path

import pytest

from .playwrightlib import playwright_setup_page_logger
from .xvirt_impl import XVirtImpl


def pytest_addoption(parser):
    parser.addoption("--headful", action="store_true", default=False, help="run tests in headfull mode")


def pytest_configure(config):
    # This function is called before any tests or fixtures are collected.
    # You can add any setup code here.

    # For example, you can add your conftest.py setup code here.
    # Note: You might need to adjust this code to work in this context.
    from .playwrightlib import playwright_patch_timeout
    playwright_patch_timeout()


@pytest.hookimpl
def pytest_xvirt_setup(config):
    headful = config.getoption("--headful")
    return XVirtImpl(headless=not headful)


def pytest_unconfigure(config):
    # This function is called after all tests and fixtures have been collected.
    # You can add any teardown code here.
    pass


@pytest.fixture(autouse=True)
def before_each_after_each(request):
    if 'page' not in request.node.fixturenames:
        yield
        return
    page = request.getfixturevalue('page')
    playwright_setup_page_logger(page)
    yield



import asyncio
from typing import Union, Coroutine, Any

# todo evaluate if best to use Awaitable instead of Coroutine
OptionalCoroutine = Union[None, Coroutine[Any, Any, None]]
_task_set = set()


def create_task_safe(coro):
    """Create an asyncio task and keep a reference to prevent garbage collection.

    See https://stackoverflow.com/questions/71938799/python-asyncio-create-task-really-need-to-keep-a-reference
    """
    task = asyncio.create_task(coro)
    _task_set.add(task)
    task.add_done_callback(_task_set.discard)
    return task



async def async_fetch_str(url: str, method: str = 'GET', data: str = '') -> str:
    pass


try:
    from wwwpy.remote.fetch import async_fetch_str
except ImportError or ModuleNotFoundError:
    from wwwpy.server.fetch import async_fetch_str



from __future__ import annotations

import logging
import sys
import time
from dataclasses import dataclass
from pathlib import Path
from typing import Collection, Sequence

from wwwpy.bootstrap import bootstrap_routes
from wwwpy.common import loglib
from wwwpy.common.rpc.custom_loader import CustomFinder
from wwwpy.common.settingslib import Settings
from wwwpy.resources import library_resources, from_directory
from wwwpy.rpc import RpcRoute
from wwwpy.server.custom_str import CustomStr
from wwwpy.webserver import Route
from wwwpy.websocket import WebsocketPool

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Config:
    """
    It may be that we need to split this in the following logic blocks:
    - dev_mode flag
    - (a set of) 'directory' has it's common/remote/server set of folders
        - this will impact the hot reload mechanism and the bundle to be sent to the remote
    - keep 'common' separate from 'remote' and 'server' folders ?
    - todo continue this analysis
    """
    directory: Path
    dev_mode: bool
    server_rpc_packages: Collection[str]
    remote_rpc_packages: Collection[str]
    server_folders: Collection[str]
    remote_folders: Collection[str]


@dataclass
class Project:
    config: Config
    settings: Settings
    websocket_pool: WebsocketPool
    routes: Sequence[Route]


def setup(config: Config, settings: Settings = None) -> Project:
    """This function is used / should be used from these points:
    - wwwpy main module
    - remote tests, i.e., XVirtImpl to setup the infrastructure to run tests
    - server tests, i.e. those using Playwright
    """
    if settings is None:
        settings = Settings()

    loglib.set_log_level(settings.log_level)
    
    directory = config.directory
    sys.path.insert(0, CustomStr(directory))
    sys.meta_path.insert(0, CustomFinder(set(config.remote_rpc_packages)))

    websocket_pool = WebsocketPool('/wwwpy/ws')

    services = _configure_server_rpc_services('/wwwpy/rpc', list(config.server_rpc_packages))
    services.generate_remote_stubs()

    resources = [library_resources(), services.remote_stub_resources(), ] + \
                [from_directory(directory / f, relative_to=directory) for f in config.remote_folders]

    routes: list[Route] = [
        services.route,
        websocket_pool.http_route,
        *bootstrap_routes(
            resources=resources,
            python=f'from wwwpy.remote.browser_main import entry_point; await entry_point(dev_mode={config.dev_mode})'
        )
    ]

    if config.dev_mode:
        import wwwpy.server.designer.dev_mode as dev_modelib
        dev_modelib._warning_on_multiple_clients(websocket_pool)

        dev_modelib.start_hotreload(
            directory, websocket_pool, services,
            server_folders=set(config.server_folders),
            remote_folders=set(config.remote_folders),
        )
        if settings.hotreload_self:
            logger.info('devself detected')
            import wwwpy
            wwwpy_dir = Path(wwwpy.__file__).parent
            wwwpy_package_dir = wwwpy_dir.parent
            dev_modelib.start_hotreload(
                wwwpy_package_dir, websocket_pool, services,
                server_folders={'wwwpy/common', 'wwwpy/server'},
                remote_folders={'wwwpy/common', 'wwwpy/remote'}
            )

    return Project(config, settings, websocket_pool, tuple(routes))


def _configure_server_rpc_services(route_path: str, modules: list[str]) -> RpcRoute:
    services = RpcRoute(route_path)
    for module_name in modules:
        services.allow(module_name)
    return services



from __future__ import annotations

import inspect
import os
from dataclasses import dataclass, field
from queue import Queue
from threading import Thread
from types import FunctionType
from typing import Callable, Any

from playwright.sync_api import PageAssertions, LocatorAssertions, APIResponseAssertions, Page, sync_playwright, \
    Playwright, Browser

with sync_playwright() as pw: pass  # workaround to run playwright in a new thread. see: https://github.com/microsoft/playwright-python/issues/1685


@dataclass
class PlaywrightBunch:
    playwright: Playwright
    page: Page
    browser: Browser


@dataclass
class PlaywrightArgs:
    url: str
    headless: bool
    queue: Queue = field(default_factory=Queue)
    instance: PlaywrightBunch | None = None


def start_playwright_in_thread(url: str, headless: bool) -> PlaywrightArgs:
    args = PlaywrightArgs(url, headless)

    def in_thread():
        start_playwright(args)
        while True:
            element = args.queue.get(timeout=30)
            if element is None:
                break
            element()

    thread = Thread(target=in_thread, daemon=True)
    thread.start()
    return args


def start_playwright(args: PlaywrightArgs) -> None:
    playwright = sync_playwright().start()
    launch_args = ['--enable-features=WebAssemblyExperimentalJSPI']
    browser = playwright.chromium.launch(headless=args.headless, args=launch_args)
    page = browser.new_page()
    playwright_setup_page_logger(page)
    page.goto(args.url)
    args.instance = PlaywrightBunch(playwright, page, browser)


def playwright_setup_page_logger(page: Page):
    page.on('console', lambda msg: print(f'console [{msg.type}] ==== {msg.text}'))
    sep = '\n' + ('=' * 60) + '\n'
    page.on('pageerror', lambda exc: print(f'{sep}uncaught exception through pageerror: {sep}{exc}{sep}'))


def playwright_patch_timeout() -> None:
    def PLAYWRIGHT_PATCH_TIMEOUT_MILLIS() -> int:
        timeout = 45000
        try:
            import wwwpy_user_conf
            timeout = wwwpy_user_conf.PLAYWRIGHT_PATCH_TIMEOUT_MILLIS
        except:
            pass
        return int(os.environ.get('PLAYWRIGHT_PATCH_TIMEOUT_MILLIS', f'{timeout}'))

    print(f'Using PLAYWRIGHT_PATCH_TIMEOUT, current value={PLAYWRIGHT_PATCH_TIMEOUT_MILLIS()}')

    # patch playwright assertion timeout to match our configuration
    # this is temporary solution until playwright supports setting custom timeout for assertions
    # github issue: https://github.com/microsoft/playwright-python/issues/1358

    def patch_timeout(_member_obj: FunctionType) -> Callable:
        def patch_timeout_inner(*args, **kwargs) -> Any:
            __tracebackhide__ = True
            timeout_millis = PLAYWRIGHT_PATCH_TIMEOUT_MILLIS()
            parameters = inspect.signature(_member_obj).parameters
            timeout_arg_index = list(parameters.keys()).index("timeout")
            if timeout_arg_index >= 0:
                if len(args) > timeout_arg_index:
                    args = list(args)  # type: ignore
                    args[timeout_arg_index] = timeout_millis  # type: ignore
                elif 'timeout' not in kwargs:
                    kwargs["timeout"] = timeout_millis
            return _member_obj(*args, **kwargs)

        return patch_timeout_inner

    for assertion_cls in [PageAssertions, LocatorAssertions, APIResponseAssertions]:
        for member_name, member_obj in inspect.getmembers(assertion_cls):
            if isinstance(member_obj, FunctionType):
                if "timeout" in inspect.signature(member_obj).parameters:
                    setattr(assertion_cls, member_name, patch_timeout(member_obj))



from __future__ import annotations

import importlib
import logging
import tempfile
import traceback
from inspect import getmembers, isfunction, signature, iscoroutinefunction, Signature
from pathlib import Path
from types import ModuleType, FunctionType
from typing import NamedTuple, List, Tuple, Optional, Callable

from wwwpy.common import modlib
from wwwpy.common.asynclib import OptionalCoroutine
from wwwpy.common.http_transport import ServerHttpTransport, RemoteHttpTransport
from wwwpy.common.rpc.hibrid_dispatcher import HybridDispatcher
from wwwpy.common.rpc.serializer import RpcRequest, RpcResponse
from wwwpy.common.rpc.v2.caller_proxy import caller_proxy_generate
from wwwpy.common.rpc2.default_skeleton import DefaultSkeleton
from wwwpy.common.rpc2.default_stub import DefaultStub
from wwwpy.common.rpc2.encoder_decoder import JsonEncoderDecoder
from wwwpy.common.rpc2.stub import generate_stub
from wwwpy.http import HttpRoute, HttpResponse, HttpRequest
from wwwpy.resources import ResourceIterable, from_directory
from wwwpy.unasync import unasync

logger = logging.getLogger(__name__)


class Function(NamedTuple):
    name: str
    func: FunctionType
    signature: str
    is_coroutine_function: bool
    sign: Signature
    blocking: FunctionType


def _std_function_to_function(fun_tuple: Tuple[str, FunctionType]) -> Function:
    name = fun_tuple[0]
    func = fun_tuple[1]
    sign = signature(func)
    is_coroutine_function = iscoroutinefunction(func)
    blocking = unasync(func)
    return Function(name, func, str(sign), is_coroutine_function, signature(func), blocking)


class SourceModule:
    def __init__(self, path: Path | str, module_name: str):
        self.path = Path(path)
        self.name = module_name

    def source(self) -> str:
        return self.path.read_text()


class Module:
    def __init__(self, module: ModuleType):
        self.module = module
        self.name = module.__name__

        self.functions: List[Function] = function_list(self.module)
        self._funcs = {f.name: f for f in self.functions}

    def __getitem__(self, name) -> Function:
        return self._funcs.get(name, None)


def function_list(module: ModuleType) -> List[Function]:
    return list(map(_std_function_to_function, getmembers(module, isfunction)))


class RpcRoute:
    def __init__(self, route_path: str):
        self._allowed_modules: set[str] = set()
        self.route = HttpRoute(route_path, self._route_callback)
        self.tmp_bundle_folder = Path(tempfile.mkdtemp())

    def _route_callback(self, request: HttpRequest,
                        resp_callback: Callable[[HttpResponse], OptionalCoroutine]) -> OptionalCoroutine:

        request_content = request.content.decode('utf-8')
        transport = ServerHttpTransport(request_content)
        encdec = JsonEncoderDecoder()
        skeleton = DefaultSkeleton(transport, encdec, self._allowed_modules)

        skeleton.invoke_tobe_fixed()

        if transport.response is None:
            raise Exception('No response was provided')

        response = HttpResponse(transport.response, 'text/plain')
        return resp_callback(response)

    def allow(self, module_name: str):
        if not isinstance(module_name, str):
            raise TypeError('module_name must be a string')
        self._allowed_modules.add(module_name)

    def find_module(self, module_name: str) -> Optional[Module]:
        if module_name not in self._allowed_modules:
            return None
        if modlib._find_module_path(module_name) is None:
            return None
        try:
            module = importlib.import_module(module_name)
        except Exception:
            logger.exception(f'Cannot import module {module_name} even though find_module_path found it')
            return None

        # todo, cache? beware of hot reload
        return Module(module)

    def find_source_module(self, module_name: str) -> Optional[SourceModule]:
        if module_name not in self._allowed_modules:
            return None
        path = modlib._find_module_path(module_name)
        if path is None:
            return None
        try:
            module = SourceModule(path, module_name)
            return module
        except Exception:
            logger.exception(f'Cannot import module {module_name} even though find_module_path found it')
            return None

    # todo could be called invoke or call_site_invoke or server_dispatch (but server is charged maybe brainstorm a name)
    #  in the rpc realm how is it called the caller and the called?
    def callee_dispatch(self, request: str) -> str:
        rpc_request = RpcRequest.from_json(request)

        module = self.find_module(rpc_request.module)
        function = module[rpc_request.func]
        exception = ''
        result = None
        try:
            result = function.blocking(*rpc_request.args)
        except Exception:
            exception = traceback.format_exc()

        response = RpcResponse(result, exception)
        return response.to_json()

    def remote_stub_resources(self) -> ResourceIterable:
        return from_directory(self.tmp_bundle_folder)

    def generate_remote_stubs(self) -> tuple[List[Path], List[Path]]:
        logger.debug(f'generate_remote_stubs in {self.tmp_bundle_folder}')
        add = []
        rem = []
        for module_name in self._allowed_modules:
            module = self.find_source_module(module_name)
            filename = module_name.replace('.', '/') + '.py'
            file = self.tmp_bundle_folder / filename
            if module is None:
                if file.exists():
                    file.unlink(missing_ok=True)
                    rem.append(file)
                continue
            module_source = module.path.read_text()
            sub_imports = '\n'.join(_make_import(o) for o in [RemoteHttpTransport, JsonEncoderDecoder]) + '\n'
            stub_args = (f'{RemoteHttpTransport.__name__}("{self.route.path}"), ' +
                         f'{JsonEncoderDecoder.__name__}(), __name__')
            stub_source = sub_imports + generate_stub(module_source, DefaultStub, stub_args)
            file.parent.mkdir(parents=True, exist_ok=True)
            file.write_text(stub_source)
            logger.debug(f'Module `{module_name}` len(stub_source)={len(stub_source)}')
            add.append(file)
        return add, rem


def generate_stub_source(module: SourceModule, rpc_url: str) -> str:
    imports = 'from wwwpy.common.fetch import async_fetch_str'
    proxy_args = f'"{module.name}", "{rpc_url}"'
    gen = caller_proxy_generate(module.source(), HybridDispatcher, proxy_args)
    gen = imports + '\n\n' + gen
    return gen


def _make_import(obj: any) -> str:
    return f'from {obj.__module__} import {obj.__name__}'



import tempfile
from pathlib import Path
from typing import List, Protocol, Any

from wwwpy.common.filesystem.sync.event import Event


def new_tmp_path() -> Path:
    return Path(tempfile.mkdtemp(prefix='debounce-tmp-path-'))

class Sync(Protocol):

    @staticmethod
    def sync_source(source: Path, events: List[Event]) -> List[Any]:
        """Could be called 'sync_produce'. This part has the following input:
- a reference to the root of the monitored source filesystem
- a list of filesystem change events


The output is:
- a list of filesystem changes that can be serialized.
- such list can be an aggregated list of changes (e.g., a file was created and then modified, the two events can be aggregated into a single event)

"""

    @staticmethod
    def sync_target(target: Path, changes: List[Any]) -> None:
        """Could be called 'sync_apply'. This part has the following input:
- a reference to the root of the target filesystem
- a list of filesystem aggregated change events

The output is:
- the target filesystem is updated to reflect the changes
"""

    @staticmethod
    def sync_init(source: Path) -> List[Any]:
        """This part has the following input:
- a reference to the root of the monitored source filesystem

The output is:
- a list of filesystem aggregated change events that reflect the current state of the source filesystem
"""



from __future__ import annotations

from abc import ABC
from typing import Callable

from . import code_edit, code_strings, html_edit, html_locator
from . import code_info
from . import element_library as el
from . import element_path as ep
from .code_edit import rename_class_attribute, add_class_attribute
from .. import modlib
from ..collectionlib import ListMap

tag_inner_html_attr_name = 'inner HTML'
tag_data_name_attr_name = 'data-name'

_ad_content_string = el.AttributeDef(tag_inner_html_attr_name,
                                     el.Help('This is the inner content of the tag', ''), )
_ad_data_name = el.AttributeDef(tag_data_name_attr_name,
                                el.Help('This is the data-name of the tag', ''), )


class AttributeEditor(ABC):
    """Allow to add/edit/remove attributes of an HTML element according to its AttributeDef."""

    @property
    def definition(self) -> el.AttributeDef:
        return self._definition

    @property
    def exists(self) -> bool:
        return self._exists

    def remove(self):
        self._remove(self)

    @property
    def value(self) -> str | None:
        return self._value

    # todo if we wanted to give a result for each operation (set, remove, etc) we could return a Result object but we cannot do it with `value` because it is a property
    @value.setter
    def value(self, value: str | None):
        self._set_value(self, value)

    def __init__(self, definition: el.AttributeDef, exists: bool, value: str | None,
                 set_value: Callable[[AttributeEditor, str | None], None],
                 remove: Callable[[AttributeEditor], None]
                 ):
        self._definition = definition  # readonly
        self._exists = exists
        self._value = value
        self._set_value = set_value
        self._remove = remove

    # definition: el.AttributeDef
    # exists: bool
    # value: str | None


class EventEditor:
    """Allow to add/edit events of an HTML element according to its EventDef.
    The removal of events is not supported."""

    @property
    def handled(self) -> bool:
        return self._handled

    @property
    def definition(self) -> el.EventDef:
        return self._definition

    @property
    def method(self) -> code_info.Method | None:
        return self._method

    def do_action(self) -> None:
        """If the class does not have the method, add it.
        In any case it should focus the IDE cursor on the method.
        """
        self._do_action(self)

    def __init__(self, event_def: el.EventDef, method: code_info.Method | None, method_name: str,
                 _do_action: Callable[[EventEditor], None]):
        self._handled = method is not None
        self._method = method
        self._definition = event_def
        self.method_name = method_name
        self._do_action = _do_action


class ElementEditor:
    """Allow to edit an HTML element inside a Component, according to the specified ElementDef.
    Notice that the changes applied by an instance of this class  are not reflected in the internal state;
    it is expected that the hot reload will be triggered and consequently this instance will be discarded and
    reloaded.
    ElementDef.attributes should not contain data-name and inner HTML attributes, they are added as attributes
    managed directly by the framework.
    """

    def __init__(self, element_path: ep.ElementPath, element_def: el.ElementDef):
        self.attributes: ListMap[AttributeEditor] = ListMap(key_func=lambda attr: attr.definition.name)
        """One AttributeEditor for each attribute defined in the ElementDef."""
        self.events: ListMap[EventEditor] = ListMap(key_func=lambda ev: ev.definition.name)
        """One EventEditor for each event defined in the ElementDef."""

        self.element_path = element_path
        self.element_def = element_def
        self._fill_events()
        self._fill_attrs()

    def _fill_events(self):
        element_def = self.element_def
        element_path = self.element_path
        ci = code_info.class_info(self.current_python_source(), element_path.class_name)
        data_name = element_path.data_name
        for event_def in element_def.events:
            method_name = f'{data_name}__{event_def.python_name}'
            method = ci.methods_by_name.get(method_name, None)
            event_editor = EventEditor(event_def, method, method_name, self._event_do_action)
            self.events.append(event_editor)

    def _fill_attrs(self):
        element_def = self.element_def
        element_path = self.element_path
        node = html_locator.locate_node(self._html_source(), element_path.path)
        self._node = node

        # Add the data-name
        exists = _ad_data_name.name in node.attributes
        value = node.attributes.get(_ad_data_name.name, None)
        attribute_editor = AttributeEditor(_ad_data_name, exists, value,
                                           self._data_name_set_value, self._attribute_remove)
        self.attributes.append(attribute_editor)

        # Add the inner HTML attribute
        if node.content_span:
            cx, cy = node.content_span
            content_exists = True if cx < cy else False
            content = node.content
            if content is not None:
                content = escape_string(content)
            self.attributes.append(AttributeEditor(_ad_content_string, content_exists, content,
                                                   self._content_string_set_value, self._content_string_set_value))

        # Add the other attributes from the ElementDef
        for attribute_def in element_def.attributes:
            exists = attribute_def.name in node.attributes
            value = node.attributes.get(attribute_def.name, None)
            attribute_editor = AttributeEditor(attribute_def, exists, value,
                                               self._attribute_set_value, self._attribute_remove)
            self.attributes.append(attribute_editor)

    def current_python_source(self):
        return self._python_source_path().read_text()

    def _python_source_path(self):
        path = modlib._find_module_path(self.element_path.class_module)
        if not path:
            raise ValueError(f'Cannot find module {self.element_path.class_module}')
        return path

    def _html_source(self):
        ps = self.current_python_source()
        html = code_strings.html_from_source(ps, self.element_path.class_name)
        return html

    def _event_do_action(self, event_editor: EventEditor):
        if event_editor.handled:
            return
        new_source = code_edit.add_method(self.current_python_source(), self.element_path.class_name,
                                          event_editor.method_name, 'event',
                                          "logger.debug(f'{inspect.currentframe().f_code.co_name} event fired %s', event)")
        self._write_source(new_source)

    def _write_source(self, new_source):
        self._python_source_path().write_text(new_source)

    def _html_change(self, html_manipulator: Callable[[str], str]):
        python_source = self.current_python_source()
        new_source = code_strings.html_string_edit(python_source, self.element_path.class_name, html_manipulator)
        self._write_source(new_source)

    def _attribute_set_value(self, attribute_editor: AttributeEditor, value: str | None):

        def _html_manipulate(html: str) -> str:
            new_html = html_edit.html_attribute_set(
                html, self.element_path.path, attribute_editor.definition.name, value)
            return new_html

        self._html_change(_html_manipulate)

    def _data_name_set_value(self, attribute_editor: AttributeEditor, value: str | None = ''):
        old_value = attribute_editor.value
        if self._node.content and old_value == self._node.content:
            # If the data-name is the same as the inner HTML, we should change the inner HTML too
            self._content_string_set_value(None, value)

        self._attribute_set_value(attribute_editor, value)
        src = self.current_python_source()
        if old_value:
            src = rename_class_attribute(src, self.element_path.class_name, old_value, value)
        else:
            # add class_attribute
            # add_class_attribute(source_code, class_name, Attribute(attr_name, comp_def.python_type, 'wpc.element()'))
            pt = self.element_def.python_type
            src = add_class_attribute(src, self.element_path.class_name,
                                      code_info.Attribute(value, pt, 'wpc.element()'))
        self._write_source(src)

    def _content_string_set_value(self, attribute_editor: AttributeEditor, value: str | None = ''):
        if value is not None:
            value = unescape_string(value)

        def _html_manipulate(html: str) -> str:
            new_html = html_edit.html_content_set(html, self.element_path.path, value)
            return new_html

        self._html_change(_html_manipulate)

    def _attribute_remove(self, attribute_editor: AttributeEditor):

        def _html_manipulate(html: str) -> str:
            new_html = html_edit.html_attribute_remove(
                html, self.element_path.path, attribute_editor.definition.name)
            return new_html

        self._html_change(_html_manipulate)


# Create a translation table for the specified characters
escape_table = str.maketrans({
    '\r': '\\r',  # Carriage return -> \r
    '\n': '\\n',  # Newline (line feed) -> \n
    '\t': '\\t',  # Tab -> \t
    '\\': '\\\\'  # Backslash -> \\
})


def escape_string(s: str) -> str:
    return s.translate(escape_table)


def unescape_string(s: str) -> str:
    escaped_bytes = s.encode('ascii', 'backslashreplace')
    return escaped_bytes.decode('unicode_escape')



import ast
from typing import Type, Optional

from wwwpy.common.rpc.v2.dispatcher import Dispatcher, Definition, FunctionDef

_annotations_type = set[ast.Name]


def caller_proxy_generate(source: str, dispatcher_callable: Type[Dispatcher], dispatcher_args: str = '') -> str:
    """
In the caller/callee this is a caller proxy generator.

    This function is used to parse a source code and generate a new source code that:
- calls definition_complete at the end of each class and the module
- handles the definition of the functions/methods with type hints
- removes the implementation of the functions and replaces it with a forwarding call to its contextual
dispatcher (being it the module or a class dispatcher)

See the protocol DispatcherBuilder
"""
    tree: ast.Module = ast.parse(source)
    module = dispatcher_callable.__module__
    qualified_name = dispatcher_callable.__qualname__
    lines = [
        f'from {module} import {qualified_name}',
        f'dispatcher = {qualified_name}({dispatcher_args})',
        ''
    ]
    used_annotations: _annotations_type = set()
    functions = {}
    for b in tree.body:
        if isinstance(b, (ast.FunctionDef, ast.AsyncFunctionDef)) and not b.name.startswith('_'):
            b.body = []  # keep only the signature
            func_def = ast.unparse(b)
            lines.append(func_def)
            args_list = []
            anno_list = []
            for ar in b.args.args:
                used_annotations.add(ar.annotation)
                args_list.append(f'{ar.arg}')
                anno_list.append(ast.unparse(ar.annotation))
            args = f'"{b.name}", ' + ', '.join(args_list)
            return_type = 'None'
            if b.returns:
                return_type = ast.unparse(b.returns)
                used_annotations.add(b.returns)
            functions[b.name] = f'FunctionDef("{b.name}", [{", ".join(anno_list)}], {return_type})'
            if isinstance(b, ast.AsyncFunctionDef):
                lines.append(f'    return await dispatcher.dispatch_async({args})')
            else:
                lines.append(f'    return dispatcher.dispatch_sync({args})')
            lines.append('')  # empty line after each function
        elif isinstance(b, (ast.ImportFrom, ast.Import)):
            lines.append(b)

    for idx in range(len(lines)):
        line = lines[idx]
        if isinstance(line, ast.Import):
            lines[idx] = ast.unparse(line) if _is_import_used(line, used_annotations) else ''
        elif isinstance(line, ast.ImportFrom):
            lines[idx] = ast.unparse(line) if _is_import_from_used(line, used_annotations) else ''

    fdict = '{' + ', '.join(f'"{fname}": {fdef}' for fname, fdef in functions.items()) + '}'

    lines.append('from ' + Definition.__module__ + ' import ' + Definition.__name__ + ', ' + FunctionDef.__name__)
    lines.append(f'dispatcher.definition_complete(Definition("module", {fdict}))')

    body = '\n'.join(lines)
    return body


def _is_import_used(node: ast.Import, used_annotations: _annotations_type) -> bool:
    for alias in node.names:
        candidate = alias.asname if alias.asname is not None else alias.name
        for ann in used_annotations:
            if _annotation_uses_candidate(ann, candidate):
                return True
    return False


def _is_import_from_used(node: ast.ImportFrom, used_annotations: _annotations_type) -> bool:
    for alias in node.names:
        candidate = alias.asname if alias.asname is not None else alias.name
        for ann in used_annotations:
            if _annotation_uses_candidate(ann, candidate):
                return True
    return False


def _annotation_uses_candidate(node: ast.AST, candidate: str) -> bool:
    full_name = _get_full_name(node)
    if full_name is not None:
        if full_name == candidate or full_name.startswith(candidate + '.'):
            return True
    for child in ast.iter_child_nodes(node):
        if _annotation_uses_candidate(child, candidate):
            return True
    return False


def _get_full_name(node: ast.AST) -> Optional[str]:
    if isinstance(node, ast.Name):
        return node.id
    elif isinstance(node, ast.Attribute):
        base = _get_full_name(node.value)
        if base is not None:
            return base + '.' + node.attr
    return None



from __future__ import annotations

import inspect
import types
import typing
from dataclasses import dataclass


@dataclass
class TypedFunction:
    module_name: str
    func_name: str
    args_types: list[type]
    return_type: type  # this MUST be present also for void functions
    is_coroutine: bool

    def __post_init__(self):
        if self.return_type is None:
            raise Exception(f'Return type missing for function {self.func_name}')


def get_typed_function(function: types.FunctionType) -> TypedFunction:
    type_hints = typing.get_type_hints(function)  # to be used if the below code do not resolve types

    signature = inspect.signature(function)
    args_types = []
    for name, param in signature.parameters.items():
        annotation = type_hints.get(name, None)
        if annotation is None:
            raise Exception(
                f'There is no support for not annotated arguments. Annotation missing for parameter {name} in function {function.__name__}')
        args_types.append(annotation)

    return_annotation = type_hints.get('return', type(None))
    return TypedFunction(
        function.__module__,
        function.__name__,
        args_types,
        return_annotation,
        inspect.iscoroutinefunction(function)
    )



import asyncio
import logging
from pathlib import Path

import js
from pyodide.ffi import create_proxy, to_js
from wwwpy.common.asynclib import create_task_safe

logger = logging.getLogger(__name__)


async def _fs_idbfs_sync(n):
    queue = asyncio.Queue(1)
    js.pyodide.FS.syncfs(n, create_proxy(lambda err: queue.put_nowait(err)))
    res = await asyncio.wait_for(queue.get(), 5)
    logger.debug(f'fs_idbfs_sync({n}) res: {res}')
    return res


def fs_idbfs_load(): return create_task_safe(_fs_idbfs_sync(1))


def fs_idbfs_save(): return create_task_safe(_fs_idbfs_sync(0))


async def fs_idbfs_mount(mount_point: str):
    if Path(mount_point).exists():
        return logger.warning('already mounted')
    js.pyodide.FS.mkdirTree(mount_point)
    js.pyodide.FS.mount(js.pyodide.FS.filesystems.IDBFS, to_js({}), mount_point)
    await fs_idbfs_load()



from pathlib import Path
from typing import List, Any

from wwwpy.common import files


class DesignerRpc:

    def hotreload_notify_changes(self, do_reload: bool, events: List[Any]):
        directory = Path(files._bundle_path)

        from wwwpy.common.filesystem.sync import Sync, sync_delta2
        sync_impl: Sync = sync_delta2
        sync_impl.sync_target(directory, events)
        if do_reload:
            self.hotreload_do()

    def hotreload_do(self):
        from wwwpy.remote.browser_main import _reload
        _reload()



import inspect
from dataclasses import dataclass

import wwwpy.remote.component as wpc
import js  # used in globals
from js import pyodide, document, console, window

import logging

from wwwpy.common import state

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


@dataclass
class State:
    python_code: str = 'logger.info("Hello")'


class PythonConsoleComponent(wpc.Component, tag_name='wwwpy-python-console'):
    _ta_python_code: js.HTMLTextAreaElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
<div>Press CTRL-Enter to run Python code</div>
<textarea data-name="_ta_python_code" rows="12" wrap="off" 
style="width: 95%;"></textarea>
"""
        self._state = state._restore(State)
        self._ta_python_code.value = self._state.python_code

    async def _ta_python_code__keydown(self, event):
        if event.ctrlKey and event.key == 'Enter':
            await js.pyodide.runPythonAsync(self._ta_python_code.value, globals=globals())

    async def _ta_python_code__input(self, event):
        self._state.python_code = self._ta_python_code.value



import json
from typing import NamedTuple, List, Any, Optional


class RpcResponse(NamedTuple):
    result: Any
    exception: str

    @classmethod
    def from_json(cls, string: str) -> 'RpcResponse':
        obj = json.loads(string)
        response = RpcResponse(*obj)
        return response

    def to_json(self) -> str:
        return json.dumps(self, default=str)


class RpcRequest(NamedTuple):
    module: str
    func: str
    args: List[Optional[Any]]

    @classmethod
    def to_json(cls, module_name: str, func_name: str, *args) -> str:
        return json.dumps(RpcRequest(module_name, func_name, args))

    @classmethod
    def from_json(cls, string: str) -> 'RpcRequest':
        obj = json.loads(string)
        request = RpcRequest(*obj)
        return request



from js import document
import js

from wwwpy.common.asynclib import create_task_safe
from wwwpy.remote.jslib import script_load_once

_version = '2.20.0'
_css_url = f'https://cdn.jsdelivr.net/npm/@shoelace-style/shoelace@{_version}/cdn/themes/dark.css'
_js_url = f'https://cdn.jsdelivr.net/npm/@shoelace-style/shoelace@{_version}/cdn/shoelace.js'

_task = []


def setup_shoelace():
    async def _setup_shoelace():
        needed = await script_load_once(_js_url, type='module')
        if needed:
            document.documentElement.className = 'sl-theme-dark'
            document.head.append(document.createRange().createContextualFragment(_head_style))

    return create_task_safe(_setup_shoelace())


# <script type="module" src="https://cdn.jsdelivr.net/npm/@shoelace-style/shoelace@2.20.0/cdn/shoelace.js" ></script>

# language=html
_head_style = f"""
<meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=0">
<style>@import '{_css_url}';</style>
<style>
    body {{
        font: 16px sans-serif;
        background-color: var(--sl-color-neutral-0);
        color: var(--sl-color-neutral-900);
        padding: 1rem;
    }}
</style>
"""
_head_script = f"""<script type="module" src="{_js_url}"></script>"""


# alternative
# language=html
_head_style_dark_palin ="""
<meta name="viewport" content="width=device-width, initial-scale=1.0, maximum-scale=1.0, user-scalable=0">
<style>
    body {
        background: #121212;
        color: #e0e0e0;
        margin: 1rem;
        font: 16px sans-serif;
    }

    a {
        color: #bb86fc
    }
</style>
"""


from __future__ import annotations

import asyncio
import logging
from typing import Dict

import js
from js import HTMLElement, console
from pyodide.ffi import create_proxy

logger = logging.getLogger(__name__)
namespace = "window.python_custom_elements"


# todo maybe we can join Metadata and Component into one class
class Metadata:
    def __init__(self, tag_name: str | None = None, clazz=None):
        if clazz is not None:
            if not issubclass(clazz, Component):
                raise Exception(f'clazz must be a subclass of {Component.__name__}')
            if tag_name is None:
                # get the full class name
                fully_qualified_class_name = clazz.__qualname__
                tag_name = ('wwwpy-auto-' + fully_qualified_class_name.lower()
                            .replace('<', '-')
                            .replace('>', '-')
                            .replace('.', '-')
                            )

        self.tag_name = tag_name
        self.observed_attributes = set()
        self.registered = False
        self.clazz = clazz
        self._custom_element_class_template = None

    def __set_name__(self, owner, name):
        if not issubclass(owner, Component):
            raise Exception(f'attribute {name} must be in a subclass of {Component.__name__}')
        self.clazz = owner
        # if self.auto_define:
        #     self.define_element()

    def define_element(self):
        if self.registered:
            return
        if js.eval(f'typeof {namespace}') == 'undefined':
            js.eval(namespace + ' = {}')

        js_class_name = self.tag_name.replace('-', '_').replace('.', '_')
        self._js_class_name = js_class_name

        pc = js.eval(namespace)
        already_defined = hasattr(pc, js_class_name)
        setattr(pc, js_class_name, self.clazz)  # set the python constructor, in any case
        if not already_defined:
            obs_attr = ', '.join(f'"{attr}"' for attr in self.observed_attributes)
            code = (_custom_element_class_template
                    .replace('$ClassName', js_class_name)
                    .replace('$tagName', self.tag_name)
                    .replace('$observedAttributes', obs_attr)
                    .replace('$namespace', namespace)
                    )
            self._custom_element_class_template = code
            # raise Exception(code)
            js.eval(code)

        self.registered = True

    @property
    def html_snippet(self):
        """Returns the complete HTML markup for this custom element."""
        return self.build_snippet()

    def build_snippet(self, attributes: Dict[str, str | None] = None, inner_html: str = '') -> str:
        """Returns a snippet of HTML for this custom element."""
        if attributes is None:
            attributes = {}
        attr_str = ' '.join(f'{k}="{v}"' for k, v in attributes.items())
        return f'<{self.tag_name} {attr_str}>{inner_html}</{self.tag_name}>'


def get_component(element: HTMLElement) -> Component | None:
    if not hasattr(element, '_python_component'):
        return None
    component = element._python_component
    if hasattr(component, "unwrap"):
        component = component.unwrap()

    return component


# PUBLIC-API
class Component:
    component_metadata: Metadata = None
    element: HTMLElement = None
    element_not_found_raises = False

    def __init_subclass__(cls, tag_name: str | None = None, auto_define=True, **kwargs):
        super().__init_subclass__(**kwargs)
        metadata = Metadata(tag_name=tag_name, clazz=cls)
        cls.component_metadata = metadata
        if metadata.clazz is None:
            metadata.clazz = cls

        for name, value in cls.__dict__.items():
            if isinstance(value, attribute):
                cls.component_metadata.observed_attributes.add(name)

        if auto_define:
            cls.component_metadata.define_element()

    def __init__(self, element_from_js=None):
        if element_from_js is None:
            self.element = js.eval(f'window.{self.component_metadata._js_class_name}').new(create_proxy(self))
        else:
            self.element = element_from_js

        self.init_component()
        self._bind_events()

        if asyncio.iscoroutinefunction(self.after_init_component):
            asyncio.create_task(self.after_init_component())
        else:
            self.after_init_component()

    async def after_init_component(self):
        """This is called after init_component, it can be async or called synchronously if it is a normal method.
        It is called after the event binding is completed"""
        pass

    def init_component(self):
        pass

    def connectedCallback(self):
        pass

    def disconnectedCallback(self):
        pass

    def adoptedCallback(self):
        pass

    def attributeChangedCallback(self, name: str, oldValue: str, newValue: str):
        pass

    def root_element(self):
        """This is used to locate the child elements"""
        e = self.element
        return e if e.shadowRoot is None else e.shadowRoot

    def _find_element(self, name: str):
        root = self.root_element()
        selector = root.querySelector(f'[data-name="{name}"]')
        if selector is None:
            msg = f'Not found data-name: [{name}] html: [{root.outerHTML}]'
            if self.element_not_found_raises:
                raise ElementNotFound(msg)
            else:
                console.warn(msg)

        return selector

    def _find_python_attribute(self, name: str):
        selector = self._find_element(name)

        if selector:
            comp = get_component(selector)
            if comp:
                selector = comp

        self._check_type(name, selector)
        return selector

    def _check_type(self, name, selector):
        import inspect
        try:
            # todo fix probably this happens after one hot reload
            # it is inspecting the annotations (wpc.HTMLElement) and it fails to
            # eval_str because it was unloaded...?
            annotations = inspect.get_annotations(self.__class__, eval_str=True)
        except Exception as e:
            logger.error(f'Error getting annotations name=`{name}` selector=`{selector}`: {e}')
            return

        expected_type = annotations.get(name, None)
        if expected_type is None:
            # raise Exception(f'No type defined for field: {name}')
            return
        # raise Exception(f'type of expected_type: {type(expected_type)}')
        # test if expected_type is a class
        if not inspect.isclass(expected_type):
            return
        if not issubclass(expected_type, Component):
            return
        # raise Exception(f'Expected type: {expected_type} for field: {name} but found: {type(selector)}')
        isinst = not isinstance(selector, expected_type)
        # raise Exception(f'isinst: {isinst}')
        if isinst:
            raise WrongTypeDefinition(f'Expected type: {expected_type} for field: {name} but found: {type(selector)}')

    def _bind_events(self):

        members = dir(self)

        for name in members:
            if name.startswith('__'):
                continue
            parts = name.split('__')
            if len(parts) != 2:
                continue

            element_name = parts[0]
            event_name = parts[1].replace('_', '-')
            if element_name == '' or event_name == '':
                continue

            element = self._find_element(element_name)
            if element is None:
                console.warn(f'Event bind failed, element `{element_name}` was not found for method `{name}`')
                continue

            m = getattr(self, name)
            element.addEventListener(event_name, create_proxy(m))


class ElementNotFound(AttributeError): pass


class WrongTypeDefinition(TypeError): pass


# language=javascript
_custom_element_class_template = """
class $ClassName extends HTMLElement {
    static observedAttributes = [ $observedAttributes ];
    constructor(python_component) {
        super();
        if (python_component) 
            this._python_component = python_component;
        else 
            this._python_component = $namespace.$ClassName(this);
    }

    connectedCallback()    { this._python_component.connectedCallback(); }
    disconnectedCallback() { this._python_component.disconnectedCallback(); }
    adoptedCallback()      { this._python_component.adoptedCallback(); }
    attributeChangedCallback(name, oldValue, newValue) { this._python_component.attributeChangedCallback(name, oldValue, newValue); }
}

customElements.define('$tagName', $ClassName);
window.$ClassName = $ClassName;
"""


# class Attribute(str):
#     _attr: attribute
#     present: bool
#
#     def toggle(self):
#         pass

# PUBLIC-API
class attribute:

    def __init__(self):
        self.name = None

    def __set_name__(self, owner, name):
        if not issubclass(owner, Component):
            raise Exception(f'attribute {name} must be in a subclass of {Component.__qualname__}')
        self.name = name

    def __get__(self, obj: Component, objtype=None):
        return obj.element.getAttribute(self.name)

    def __set__(self, obj: Component, value):
        obj.element.setAttribute(self.name, value)


# PUBLIC-API
class element:

    def __init__(self):
        self.name = None

    def __set_name__(self, owner, name):
        if not issubclass(owner, Component):
            raise Exception(f'attribute {name} must be in a subclass of {Component.__qualname__}')
        self.name = name

    def __get__(self, obj: Component, objtype=None):
        return obj._find_python_attribute(self.name)



from __future__ import annotations
from dataclasses import dataclass


@dataclass(frozen=True)
class Package:
    name: str
    version: str


@dataclass(frozen=True)
class PackageSpecification:
    name: str
    version_specs: str | None = None
    """Version specifications for the package or None for any version. E.g. '==1.0.0' or '>=1.0.0'
        
    See https://packaging.python.org/en/latest/specifications/version-specifiers/
    """

    def build_installation_string(self) -> list[str]:
        if self.version_specs is None:
            return [self.name]
        return [f"{self.name}{self.version_specs}"]


@dataclass(frozen=True)
class PackageRequest:
    install: list[PackageSpecification]



import asyncio
from hypercorn.config import Config
from hypercorn.asyncio import serve

from echo_handler import app

# uv pip install hypercorn
config = Config()
config.bind = '0.0.0.0:8000'
asyncio.run(serve(app, config))



from __future__ import annotations

import time
from pathlib import Path

from wwwpy.common import quickstart
from wwwpy.common.designer import log_emit
from wwwpy.server import tcp_port
from wwwpy.server.configure import Project, logger, Config, setup
from wwwpy.server.settingslib import user_settings
from wwwpy.webservers.available_webservers import available_webservers

_projects: list[Project] = []


def default_project() -> Project:
    if len(_projects) == 0:
        raise ValueError('The default project has not been set')
    return _projects[0]


def add_project(project: Project):
    _projects.append(project)


def start_default(directory: Path, port: int, dev_mode=False) -> Project:
    quickstart.warn_if_unlikely_project(directory)

    config = default_config(directory, dev_mode)
    project = setup(config, user_settings())
    add_project(project)

    webserver = available_webservers().new_instance()
    webserver.set_routes(*project.routes)

    while tcp_port.is_port_busy(port):
        logger.warning(f'port {port} is busy, retrying...')
        [time.sleep(0.1) for _ in range(20) if tcp_port.is_port_busy(port)]

    webserver.set_port(port).start_listen()

    return project


def default_config(directory: Path, dev_mode: bool) -> Config:
    server_rpc_packages = ['server.rpc']
    remote_rpc_packages = {'remote', 'remote.rpc', 'wwwpy.remote', 'wwwpy.remote.rpc'}
    server_folders = {'common', 'server'}
    remote_folders = {'common', 'remote'}

    if dev_mode:
        server_rpc_packages.append('wwwpy.server.designer.rpc')
        remote_rpc_packages.update({'wwwpy.remote.designer', 'wwwpy.remote.designer.rpc'})
        log_emit.add_once(print)

    return Config(
        directory=directory,
        dev_mode=dev_mode,
        server_rpc_packages=server_rpc_packages,
        remote_rpc_packages=remote_rpc_packages,
        server_folders=server_folders,
        remote_folders=remote_folders
    )



from __future__ import annotations

import asyncio
from dataclasses import dataclass
from enum import Enum
from typing import Optional, List

import js
import wwwpy.remote.component as wpc
from pyodide.ffi import create_proxy
from wwwpy.common import state, property_monitor
from wwwpy.common.designer import element_library, html_locator, code_strings
from wwwpy.common.designer.element_editor import ElementEditor, EventEditor
from wwwpy.common.designer.element_path import ElementPath, Origin
from wwwpy.remote import dict_to_js
from wwwpy.remote.designer.helpers import _element_path_lbl, _rpc_save, _log_event, _help_button

from .button_tab import ButtonTab, Tab
from .help_icon import HelpIcon
from .searchable_combobox2 import SearchableComboBox, Option


# write enum class with [events, attributes] and use it in the button click event
class PETab(str, Enum):
    """Property Editor Tab"""
    palette = 'palette'
    attributes = 'attributes'
    events = 'events'


@dataclass
class PropertyEditorState:
    mode: PETab = PETab.attributes


class PropertyEditor(wpc.Component, tag_name='wwwpy-property-editor'):
    row_container: js.HTMLElement = wpc.element()
    message1div: js.HTMLElement = wpc.element()
    _tabs: ButtonTab = wpc.element()
    _selected_element_path: Optional[ElementPath] = None
    state: PropertyEditorState = PropertyEditorState()

    @property
    def selected_element_path(self) -> Optional[ElementPath]:
        return self._selected_element_path

    @selected_element_path.setter
    def selected_element_path(self, value: Optional[ElementPath]):
        self._selected_element_path = value
        self._render()

    def init_component(self):
        self._state_manager = state.State(state.JsStorage(), 'wwwpy.toolbar.property-editor.state')
        self.state = self._state_manager.restore(PropertyEditorState).instance_or_default()
        property_monitor.monitor_changes(self.state, lambda *a: self._state_manager.save(self.state))

        # language=html
        self.element.innerHTML = """
<style>
        .wwwpy-property-editor {
            display: grid;
            grid-template-columns: 2fr 3fr;
            box-sizing: content-box;
        }
        .wwwpy-property-editor-row {
            display: contents;
        }
        .wwwpy-property-editor-row > :first-child {
            display: flex;
            align-items: center;
            padding-left: 5px;
            color: #e0e0e0;
            font-family: Arial, sans-serif;
        }
        .wwwpy-pe-box-padding {
            padding: 3px;
        }
        .wwwpy-pe-event-value {
             width: 90%;
            margin: 2px 0;
            padding: 5px;
            background-color: #2a2a2a;
            color: #e0e0e0;
            border: 1px solid #444;
            border-radius: 4px;
            font-size: 14px;
        }
    </style>
<div data-name="message1div" style='color: white; margin: 0.3em'>&nbsp</div>
<wwwpy-button-tab data-name="_tabs"></wwwpy-button-tab>
<div  data-name='row_container' class="wwwpy-property-editor"></div>

        """

        def set_state_render(mode: PETab):
            self.state.mode = mode
            self._render()

        def new_tab(petab: PETab):
            return Tab(petab.name, on_selected=lambda tab: set_state_render(petab))

        self._tab_palette = new_tab(PETab.palette)
        self._tab_attributes = new_tab(PETab.attributes)
        self._tab_events = new_tab(PETab.events)
        self._tabs.tabs = [self._tab_palette, self._tab_attributes, self._tab_events, ]

    def set_state_selection_active(self):
        if self.state.mode == PETab.palette:
            self.state.mode = PETab.attributes

    def add_row(self, row: wpc.Component):
        row.element.classList.add('wwwpy-property-editor-row')
        self.row_container.appendChild(row.element)

    def _render(self):
        ep = self._selected_element_path
        # this is done here (and not in the moment of saving the ep) because
        # the stored ep could change in any moment, for example the user
        # change the type of tag from sl-button to sl-switch
        ep = _rebase_element_path_to_origin_source(ep)
        self.message1div.innerHTML = '' if ep is None else f'Selection: {_element_path_lbl(ep)}'

        self.row_container.innerHTML = ''
        visible = 'hidden' if ep is None else 'visible'
        self._tab_events.root_element().style.visibility = visible
        self._tab_attributes.root_element().style.visibility = visible
        if not ep:
            self._tab_palette.selected = True
            return
        lib = element_library.element_library()
        element_def = lib.by_tag_name(ep.tag_name)
        if not element_def:
            self.message1div.innerHTML += '<br>No metadata found for editing.'
            return
        self.message1div.innerHTML = f'Selection: {_element_path_lbl(ep)} {_help_button(element_def)}'
        html = code_strings.html_from(ep.class_module, ep.class_name)
        if not html:
            self.message1div.innerHTML += '<br>No HTML found for editing.'
            return

        cst_node = html_locator.locate_node(html, ep.path)
        if cst_node is None:
            self.message1div.innerHTML += '<br>No element found in the HTML.'
            return

        self.row_container.innerHTML = ''
        if self.state.mode == PETab.attributes:
            self._render_attribute_editor(element_def, ep)
            self._tab_attributes.selected = True
        elif self.state.mode == PETab.events:
            self._render_event_editor(element_def, ep)
            self._tab_events.selected = True
        else:
            self._tab_palette.selected = True

    def _set_title(self, lbl, value):
        pe_title = PE_title_row()
        pe_title.label.innerHTML = lbl
        pe_title.value.innerHTML = value
        self.add_row(pe_title)

    def _render_attribute_editor(self, element_def, ep):
        self._set_title('Attribute', 'Value')
        element_editor = ElementEditor(ep, element_def)
        for attr_editor in element_editor.attributes:
            attr_def = attr_editor.definition
            row1 = PE_attribute() if not attr_def.boolean else PE_attribute_bool()
            self.add_row(row1)

            row1.label.label.innerHTML = attr_def.name
            row1.label.set_help(attr_def.help)

            if not attr_def.boolean:
                options: List[Option] = [Option(value) for value in attr_def.values]
                focus_search = len(options) > 0
                for option in options:
                    if option.text == '':
                        option.label = 'Set to empty string'
                        option.italic = True

                if attr_editor.exists:
                    if not attr_def.mandatory:
                        remove = Option('remove attribute')
                        remove.actions.set_input_value = False
                        remove.italic = True

                        def remove_selected(ae=attr_editor):
                            ae.remove()
                            self._save(element_editor)

                        remove.on_selected = remove_selected  # lambda ae=attr_editor: ae.remove()
                        options.insert(0, remove)
                click_for = 'Click for options...'
                if attr_def.boolean:
                    row1.value.placeholder = 'attribute present' if attr_editor.exists else click_for
                else:
                    pre_placeholder = 'Set to empty string. ' if attr_editor.exists and attr_editor.value == '' else ''
                    row1.value.placeholder = pre_placeholder + (click_for if len(options) > 0 else '')
                row1.value.option_popup.options = options
                row1.value.text_value = '' if attr_editor.value is None else attr_editor.value
                row1.value.option_popup.search_placeholder = 'Search options...'
                row1.value.focus_search_on_popup = focus_search

                def attr_changed(event, ae=attr_editor, row=row1):
                    js.console.log(f'attr_changed {ae.definition.name} {row.value.text_value}')
                    ae.value = row.value.text_value
                    self._save(element_editor)

                row1.value.element.addEventListener('wp-change', create_proxy(attr_changed))
            else:
                if attr_editor.exists:
                    row1.value.checked = True
                else:
                    row1.value.checked = False

                def attr_changed(event, ae=attr_editor, row=row1):
                    if row.value.checked:
                        ae.value = None  # beware, this is creating the attribute with no value
                    else:
                        ae.remove()
                    self._save(element_editor)

                row1.value.addEventListener('change', create_proxy(attr_changed))

    def _save(self, element_editor):
        source = element_editor.current_python_source()

        async def start_save():
            await _rpc_save(self._selected_element_path, source)

        asyncio.create_task(start_save())

    def _render_event_editor(self, element_def, ep):
        self._set_title('Event', 'Value')
        element_editor = ElementEditor(ep, element_def)
        for event_editor in element_editor.events:
            row1 = PE_event()
            self.add_row(row1)
            row1.label.label.innerHTML = event_editor.definition.name
            row1.label.set_help(event_editor.definition.help)
            row1.value.placeholder = '' if event_editor.handled else 'Double click creates handler'
            row1.value.readOnly = True
            row1.value.value = event_editor.method.name if event_editor.handled else ''

            def dblclick(ev: EventEditor = event_editor):
                js.console.log(f'dblclick on {ev.definition.name}')
                handled = ev.handled
                ev.do_action()
                source = element_editor.current_python_source()

                async def start_save():
                    if not handled:
                        await _rpc_save(ep, source)
                    await _log_event(element_editor, ev)

                asyncio.create_task(start_save())

            row1.double_click_handler = lambda ev=event_editor: dblclick(ev)


class PropertyEditorRow(wpc.Component, tag_name='wwwpy-property-editor-row'):
    label: js.HTMLElement = wpc.element()
    value: js.HTMLElement = wpc.element()

    def init_component(self):
        self.element.attachShadow(dict_to_js({'mode': 'open'}))
        # language=html
        self.element.shadowRoot.innerHTML = """
        <slot name="label" data-name="label"></slot>
        <slot name="value" data-name="value"></slot>
            """


class PE_label(wpc.Component, tag_name='wwwpy-pe-label'):
    label: js.HTMLElement = wpc.element()
    _help: HelpIcon = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
        <div style="width: 100%; display: flex; justify-content: space-between;">
            <span data-name="label"></span>
            <wwwpy-help-icon data-name="_help"></wwwpy-help-icon>
        </div>
            """

    def set_help(self, help: element_library.Help):
        self._help.href = help.url
        self._help.visible = help.url != ''


class PE_event(wpc.Component):
    label: PE_label = wpc.element()
    value: js.HTMLInputElement = wpc.element()
    double_click_handler = None

    def init_component(self):
        # language=html
        self.element.innerHTML = """
        <wwwpy-pe-label data-name="label"></wwwpy-pe-label> 
        <div>
            <input data-name='value' type="text" class="wwwpy-pe-box-padding wwwpy-pe-event-value">
        </div>
            """

    def value__dblclick(self, event):
        if self.double_click_handler:
            self.double_click_handler()


class PE_attribute(wpc.Component):
    label: PE_label = wpc.element()
    value: SearchableComboBox = wpc.element()
    double_click_handler = None

    def init_component(self):
        # language=html
        self.element.innerHTML = """       
        <wwwpy-pe-label data-name="label"></wwwpy-pe-label>        
        <wwwpy-searchable-combobox2 data-name='value' class="wwwpy-pe-box-padding"></wwwpy-searchable-combobox2>
            """

    def value__dblclick(self, event):
        if self.double_click_handler:
            self.double_click_handler()


class PE_attribute_bool(wpc.Component):
    label: PE_label = wpc.element()
    value: js.HTMLInputElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
        <wwwpy-pe-label data-name="label"></wwwpy-pe-label>
        <div>
            <input type="checkbox" data-name="value" style="transform: scale(1.4)">
        </div>
            """


class PE_title_row(wpc.Component):
    label: js.HTMLElement = wpc.element()
    value: js.HTMLElement = wpc.element()

    def init_component(self):
        # language=html
        self.element.innerHTML = """
        <div style='font-weight: bold' data-name="label">Event</div><div style='font-weight: bold' data-name='value'>Value</div></div>        
            """


def _rebase_element_path_to_origin_source(ep: ElementPath) -> Optional[ElementPath]:
    """This is similar to rebase_path dumb because we use indexes alone.
    This rebase from Origin.live to Origin.source
    """
    if not ep:
        return None
    if ep.origin == Origin.source:
        return ep

    html = code_strings.html_from(ep.class_module, ep.class_name)
    if not html:
        return ep

    cst_node = html_locator.locate_node(html, ep.path)
    if cst_node is None:
        return ep

    node_path = html_locator.node_path_from_leaf(cst_node)
    ep_source = ElementPath(ep.class_module, ep.class_name, node_path, Origin.source)
    return ep_source



import asyncio

import js
from js import document
from pyodide.ffi import create_proxy


async def script_load_once(src: str, type=None, **kwargs) -> bool:
    """Returns True if the script was needed, False if it was not needed."""
    # logger.debug(f'load {src}')
    src_short = src.split('/')[-1]
    scripts = list(filter(lambda s: s.src == src, document.head.querySelectorAll('script')))  # noqa
    scripts: list[js.HTMLScriptElement]
    need_load = not scripts
    script = None if need_load else scripts[0]

    if script:
        # logger.debug(f'  found script with same src: {src_short} type={script.type}')
        assert script.type == type
    else:
        # logger.debug(f'  load is needed, no script with same src found: {src_short}')
        script = document.createElement('script')
        # copy all kwargs to the script
        for key, value in kwargs.items():
            setattr(script, key, value)
        script.src = src
        if type:
            script.type = type
        script.__async_event = asyncio.Event()
        script.onload = create_proxy(lambda e: script.__async_event.set())
        document.head.append(script)

    # logger.debug(f'  waiting for script to load: {src_short}')
    await script.__async_event.wait()
    # logger.debug(f'  event set: {src_short}')
    return need_load
